import re
from typing import Any, Callable, Dict, List, Optional, Union

import sqlalchemy as sa
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .utils import extract_table


def to_safe_snake_case(name: str) -> str:
    """Convert `name` to snake case and also add a leading underscore to variables starting with a digit.

    Args:
        name (str): The name to convert.

    Returns:
        str: The converted name.
    """
    # remove trailing space.
    name = name.strip()
    # convert space to underscores.
    name = re.sub(r"\s+", "_", name)
    # convert camel case to underscores.
    if not name.isupper():
        name = re.sub(
            r"([a-z0-8])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", name
        )
    # varable names can't start with number, so add leading underscore.
    if re.match(r"^\d", name):
        name = f"_{name}"
    # make name lowercase.
    return name.lower()


class ColumnNameFilter:
    def __init__(
        self,
        sa_obj: Union[sa.Table, DeclarativeMeta],
        column_name_map: Optional[Dict[str, str]] = None,
        column_name_converter: Callable = None,
        column_name_converter_map: Optional[Dict[str, Callable]] = None,
        columns_no_load: Optional[List[str]] = None,
    ):
        """Filter columns and convert to match database column names.

        Args:
            sa_obj (Union[sa.Table, DeclarativeMeta]): The SQLAlchemy table or entity that this filter is for.
            column_name_map (Optional[Dict[str, str]], optional): Map name to desired name. Defaults to None.
            column_name_converter (Callable, optional): A formatting function to apply to every name. Defaults to None.
            column_name_converter_map (Optional[Dict[str, Callable]], optional): Map name to formatting function. Defaults to None.
            columns_no_load (Optional[List[str]], optional): Names that should be filtered from rows. Defaults to None.
        """
        self.table_column_names = {str(c) for c in extract_table(sa_obj).columns.keys()}
        self.column_name_map = column_name_map or {}
        self.column_name_converter = column_name_converter
        self.column_name_converter_map = column_name_converter_map or {}
        self.columns_no_load = columns_no_load or []
        self._converted_column_names: Dict[str, str] = {}

    def filter_row(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """Filter unwanted columns from `row` and convert names to database column names.

        Args:
            row (Dict[str, Any]): The row to be filtered.

        Returns:
            Dict[str, Any]: The filtered row.
        """
        row = {
            self._convert_name(c): v
            for c, v in row.items()
            if c not in self.columns_no_load
        }
        # remove columns not belonging to this table.
        row = {c: v for c, v in row.items() if c in self.table_column_names}
        return row

    def _convert_name(self, name: str) -> str:
        """Convert `name` to a table column name.

        Args:
            name (str): The Name that should be converted.

        Returns:
            str: The converted name.
        """
        if name in self._converted_column_names:
            return self._converted_column_names[name]
        orig_name = name
        if name in self.column_name_map:
            name = self.column_name_map[name]
        elif name in self.column_name_converter_map:
            name = self.column_name_converter_map[name](name)
        elif self.column_name_converter:
            name = self.column_name_converter(name)
        self._converted_column_names[orig_name] = name
        return name
