from logging import Logger
from typing import Any, Callable, Dict, List, Optional, Union

import sqlalchemy as sa
from logger_cfg import logger
from sqlalchemy.orm.decl_api import DeclarativeMeta


class ColumnValueFilter:
    def __init__(
        self,
        sa_obj: Union[sa.Table, DeclarativeMeta],
        column_value_map: Optional[Dict[str, Dict[str, Any]]] = None,
        column_value_converter: Optional[Callable] = None,
        column_value_converter_map: Optional[Dict[str, Callable]] = None,
        values_no_load: List[str] = [""],
        drop_on_type_conversion_fail: bool = True,
        logger: Logger = logger,
    ) -> None:
        """Format/convert column values.

        Args:
            sa_obj (Union[sa.Table, DeclarativeMeta]): The SQLAlchemy table or entity that the filter is for.
            column_value_map (Optional[Dict[str, Dict[str, Any]]], optional): Map value to desired value. Defaults to None.
            column_value_converter (Optional[Callable], optional): A conversion function to apply to every value. Defaults to None.
            column_value_converter_map (Optional[Dict[str, Callable]], optional): Map column name to column value conversion function. Defaults to None.
            values_no_load (Optional[List[str]], optional): Values where columns should be dropped if they are that value. (e.g ['null', None, 'Na']). Defaults to None.
            drop_on_type_conversion_fail (bool, optional): Remove column from row if conversion fails. Defaults to False.
        """
        self.sa_obj = sa_obj
        self.column_value_map = column_value_map or {}
        self.column_value_converter = column_value_converter
        self.column_value_converter_map = column_value_converter_map or {}
        self.values_no_load = values_no_load or []
        self.drop_on_type_conversion_fail = drop_on_type_conversion_fail
        self.logger = logger

    def filter_row(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """Filter unwanted columns from `row` and convert values.
        The first of (column_value_map, column_value_converter_map, column_value_converter, standard type cast) that is applicable to a column will be applied to that column.

        Args:
            row (Dict[str, Any]): The row to be filtered.

        Returns:
            Dict[str, Any]: The filtered row.
        """
        # Clean, format, convert column values. (this should be called after clean_column_names)
        if self.values_no_load:
            row = {
                col: val for col, val in row.items() if val not in self.values_no_load
            }
        for col, val in list(row.items()):
            if val in self.column_value_map:
                row[col] = self.column_value_map[val]
            else:
                try:
                    if col in self.column_value_converter_map:
                        row[col] = self.column_value_converter_map[col](val)
                    elif self.column_value_converter:
                        row[col] = self.column_value_converter(val)
                except Exception as e:
                    msg = f"Error converting value {row[col]}: {e}"
                    if self.drop_on_type_conversion_fail:
                        self.logger.error(msg)
                        val = row.pop(col)
                        self.logger.warning(f"Removed value from column {col}: {val}")
                    else:
                        raise type(e)(msg)
        return row
