from typing import Union

import sqlalchemy as sa
from logger_cfg import logger
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.asyncio.engine import AsyncConnection
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .types import HypertablePartiton
from .utils import extract_table, get_typed_columns


async def create_table(
    conn: AsyncConnection,
    create_from: Union[sa.Table, DeclarativeMeta],
    recreate: bool = False,
) -> bool:
    """Create a table in the database.

    Args:
        conn (AsyncConnection): AsyncConnection to use for executing SQL statements.
        create_from (Union[sa.Table, DeclarativeMeta]): The entity or table object to create a database table for.
        recreate (bool): If table exists, drop it and recreate. Defaults to False.

    Returns:
        bool: True if a table was created, else False.
    """
    table = extract_table(create_from)
    schema = table.schema or "public"
    # create schema if needed.
    if not await conn.run_sync(conn.dialect.has_schema, schema=schema):
        logger.info(f"Creating schema '{schema}'")
        await conn.execute(sa.schema.CreateSchema(schema))
    schema_table = f"{schema}.{table.name}"
    # check if table exists.
    if await conn.run_sync(
        conn.dialect.has_table, table_name=table.name, schema=schema
    ):
        if not recreate:
            return False
        logger.info(f"Dropping table {schema_table}")
        await conn.run_sync(table.drop)
    # make sure any new Enum types are created.
    await create_enum_types(conn, table)
    logger.info(f"Creating table {schema_table}")
    await conn.run_sync(table.create)
    # create hypertable partition if needed.
    await create_hypertable_partition(conn, create_from)
    return True


async def create_enum_types(
    conn: AsyncConnection, sa_obj: Union[sa.Table, DeclarativeMeta]
) -> None:
    """Create enum types if they do not already exist in the database.
    This is necessary because sa.Table.create has no way to handle columns that have an enum type that already exists in the Database.

    Args:
        conn (AsyncConnection): AsyncConnection to use for executing SQL statements.
        sa_obj (Union[sa.Table, DeclarativeMeta]): The table who's enums should be checked.
    """
    table = extract_table(sa_obj)
    # all enums present in table.
    enums = [
        col.type
        for col in table.columns.values()
        if isinstance(col.type, postgresql.ENUM)
    ]
    for enum in enums:
        # check if enum type exists, create it if it doesn't.
        await conn.run_sync(enum.create, checkfirst=True)
        # prevent table.create for attempting to create the enum type.
        enum.create_type = False


async def create_hypertable_partition(
    conn: AsyncConnection, sa_obj: Union[sa.Table, DeclarativeMeta]
) -> None:
    """Create a hypertable partition if table or entity has a column that is flagged as a hypertable partition column.

    Args:
        conn (AsyncConnection): AsyncConnection to use for executing SQL statements.
        sa_obj (Union[sa.Table, DeclarativeMeta]): The entity or table that should be checked.
    """
    table = extract_table(sa_obj)
    if len(
        hypertable_partiton_columns := get_typed_columns(HypertablePartiton, sa_obj)
    ):
        logger.info(
            f"Creating hypertable partition using column {hypertable_partiton_columns}."
        )

        await conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS timescaledb"))
        schema = table.schema or "public"
        statement = sa.select(
            sa.text(
                f"create_hypertable('{schema}.{table.name}', '{hypertable_partiton_columns[0]}')"
            )
        )
        await conn.execute(statement)
