import os
from uuid import uuid4

import pytest
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine


@pytest.fixture
async def engine(event_loop) -> AsyncEngine:
    # connection can't be opened before `event_loop` is initialized or it will bind to the wrong loop.
    # engine format: postgresql+asyncpg://{user}:{password}@{host}:{port}/{db}
    e = create_async_engine(os.environ["DB_URL"])
    yield e
    await e.dispose()


@pytest.fixture
async def testing_schema(engine) -> str:
    schema_name = f"_testing"
    async with engine.begin() as conn:
        if not await conn.run_sync(engine.dialect.has_schema, schema=schema_name):
            await conn.execute(sa.schema.CreateSchema(schema_name))
    yield schema_name
    async with engine.begin() as conn:
        await conn.execute(sa.schema.DropSchema(schema_name, cascade=True))


@pytest.fixture
def table():
    return sa.Table(
        str(uuid4()),
        sa.MetaData(),
        sa.Column("column", sa.String),
    )


@pytest.fixture
async def keyed_table(engine, testing_schema):
    table = sa.Table(
        str(uuid4()),
        sa.MetaData(schema=testing_schema),
        # use non-nullable column.
        sa.Column("key", sa.String, primary_key=True),
        sa.Column("data", sa.String, nullable=False),
    )
    async with engine.begin() as conn:
        await conn.run_sync(table.create)
    yield table
    async with engine.begin() as conn:
        await conn.execute(table.delete())


@pytest.fixture
async def two_column_keyed_table(engine, testing_schema):
    table = sa.Table(
        str(uuid4()),
        sa.MetaData(schema=testing_schema),
        # use non-nullable column.
        sa.Column("key", sa.String, primary_key=True),
        sa.Column("data1", sa.String, nullable=False),
        sa.Column("data2", sa.String, nullable=False),
    )
    async with engine.begin() as conn:
        await conn.run_sync(table.create)
    yield table
    async with engine.begin() as conn:
        await conn.execute(table.delete())
