from copy import deepcopy
from uuid import uuid4

import pytest
import sqlalchemy as sa
from database_utils.upsert import Upserter

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


def random_rows(table: sa.Table, n_rows: int):
    column_names = [str(c) for c in table.columns.keys()]
    return [{c: str(uuid4()) for c in column_names} for _ in range(n_rows)]


def row_count_query(table):
    return sa.select(sa.func.count(sa.text("*"))).select_from(table)


async def test_rows_load(keyed_table, engine):
    """Check that rows load to database."""
    upserter = await Upserter.create(keyed_table, engine=engine)
    n_rows = 10
    await upserter.upsert(random_rows(keyed_table, n_rows))
    async with engine.begin() as conn:
        res = await conn.execute(row_count_query(keyed_table))
        db_row_count = res.scalar()
    assert db_row_count == n_rows


async def test_on_dupe_ignore(keyed_table, engine):
    upserter = await Upserter.create(keyed_table, on_dupe_update=False, engine=engine)

    n_rows = 10
    rows = random_rows(keyed_table, n_rows)

    # load rows.
    await upserter.upsert(rows)

    n_pkey_to_update = 5
    for row in rows[:n_pkey_to_update]:
        # rows with new key should get inserted.
        row["key"] += str(uuid4())
    for row in rows[n_pkey_to_update:]:
        # rows with same key should be ignore regardless of updated data.
        row["data"] += str(uuid4())
    # check that duplicate rows get ignored without error and new rows get inserted.
    await upserter.upsert(rows)
    async with engine.begin() as conn:
        res = await conn.execute(row_count_query(keyed_table))
        db_row_count = res.scalar()
    assert db_row_count == n_rows + n_pkey_to_update
    local_values_updated = {r["data"] for r in rows[n_pkey_to_update:]}
    local_values_updated_keys = [r["key"] for r in rows[n_pkey_to_update:]]
    # check that duplicate key rows with values that were updated locally do not have their values updated in the database.
    query = sa.select(keyed_table.c.data).where(
        keyed_table.c.key.in_(local_values_updated_keys)
    )
    async with engine.begin() as conn:
        res = await conn.execute(query)
        db_values = set(res.scalars())

    assert not len(db_values.intersection(local_values_updated))


async def test_on_dupe_update_column(two_column_keyed_table, engine):
    upserter = await Upserter.create(
        two_column_keyed_table, on_dupe_update=["data1"], engine=engine
    )

    n_rows = 5
    rows = random_rows(two_column_keyed_table, n_rows)

    # load original rows.
    await upserter.upsert(rows)

    # update row values.
    updated_rows = deepcopy(rows)
    for row in updated_rows:
        row["data1"] += str(uuid4())
        row["data2"] += str(uuid4())

    # load updated rows.
    await upserter.upsert(updated_rows)

    # check that all rows in database have data1 value updated.
    async with engine.begin() as conn:
        res = await conn.execute(sa.select(two_column_keyed_table.c.data1))
        db_data1 = set(res.scalars())
    assert not len(db_data1.intersection([r["data1"] for r in rows]))

    # check that no rows in database have data2 value updated.
    async with engine.begin() as conn:
        res = await conn.execute(sa.select(two_column_keyed_table.c.data2))
        db_data2 = set(res.scalars())
    assert not len(db_data2.symmetric_difference([r["data2"] for r in rows]))


async def test_on_dupe_update(two_column_keyed_table, engine):
    upserter = await Upserter.create(
        two_column_keyed_table, on_dupe_update=True, engine=engine
    )

    n_rows = 5
    rows = random_rows(two_column_keyed_table, n_rows)

    # load original rows.
    await upserter.upsert(rows)

    # update row values.
    updated_rows = deepcopy(rows)
    for row in updated_rows:
        row["data1"] += str(uuid4())
        row["data2"] += str(uuid4())

    # load updated rows.
    await upserter.upsert(updated_rows)

    # check that all rows in database have data1 value updated.
    async with engine.begin() as conn:
        res = await conn.execute(sa.select(two_column_keyed_table.c.data1))
        db_data1 = set(res.scalars())
    assert not len(db_data1.intersection([r["data1"] for r in rows]))

    # check that all rows in database have data2 value updated.
    async with engine.begin() as conn:
        res = await conn.execute(sa.select(two_column_keyed_table.c.data2))
        db_data2 = set(res.scalars())
    assert not len(db_data2.intersection([r["data2"] for r in rows]))


async def test_remove_duplicates(keyed_table, engine):
    rows = random_rows(keyed_table, 5)

    n_dupes = 2
    rows += rows[:n_dupes]

    upserter = await Upserter.create(keyed_table, remove_duplicates=True, engine=engine)
    filtered_rows = upserter._remove_duplicates(rows)
    assert len(filtered_rows) == len(rows) - n_dupes


async def test_batch_by_columns_present(two_column_keyed_table, engine):
    upserter = await Upserter.create(
        two_column_keyed_table, batch_by_columns_present=True, engine=engine
    )
    rows = [
        {"key": str(uuid4()), "data1": str(uuid4()), "data2": str(uuid4())},
        {"key": str(uuid4()), "data1": str(uuid4())},
    ]
    groups = upserter._groupby_columns(rows)
    assert len(groups) == 2
    assert isinstance(groups[0], list)
    assert isinstance(groups[1], list)
