from datetime import datetime
from uuid import uuid4

import pytest
import sqlalchemy as sa
from database_utils.upsert import Upserter
from database_utils.upsert_sources import SourcesUpserter

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
async def sources_linked_table(engine, testing_schema):
    table = sa.Table(
        str(uuid4()),
        sa.MetaData(schema=testing_schema),
        sa.Column("data_id", sa.String, primary_key=True),
        sa.Column("column1", sa.String, comment="DataIDIgnore"),
        sa.Column("column2", sa.String),
    )
    async with engine.begin() as conn:
        await conn.run_sync(table.create)
    yield table
    async with engine.begin() as conn:
        await conn.run_sync(table.drop)


@pytest.fixture
async def sources_upserter(engine, sources_linked_table):
    upserter = await SourcesUpserter.create(
        source="test",
        upserter=await Upserter.create(sources_linked_table, engine=engine),
    )
    yield upserter
    async with engine.begin() as conn:
        await conn.run_sync(upserter._sources_table.drop)


async def test_upsert(
    engine,
    sources_upserter,
):
    row = {"column2": str(uuid4())}
    query = sa.select(
        [
            sources_upserter._sources_table.c.last_seen,
            sources_upserter._sources_table.c.times_seen,
        ]
    ).where(sources_upserter._sources_table.c.source == "test")
    for i in range(1, 3):
        t_before_upsert = datetime.utcnow()
        await sources_upserter.upsert([row])
        async with engine.begin() as conn:
            res = await conn.execute(query)
            rows = res.fetchall()
        # check that same row was updated instead of inserting new.
        assert len(rows) == 1
        last_seen, times_seen = rows.pop()
        assert last_seen > t_before_upsert
        assert times_seen == i


async def test_data_id_columns(sources_upserter, engine):
    assert len(sources_upserter._data_id_columns) == 1
    assert sources_upserter._data_id_columns[0] == "column2"


async def test_create_row_data_id(sources_upserter, engine):
    _id = sources_upserter._create_row_data_id({"column2": str(uuid4())})
    assert isinstance(_id, str)
    assert len(_id)
