from collections import defaultdict
from logging import Logger
from typing import Any, Dict, List, Optional, Union

import sqlalchemy as sa
from logger_cfg import logger
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.dml import Insert
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .column_names import ColumnNameFilter
from .column_values import ColumnValueFilter
from .create import create_table
from .utils import MAX_QUERY_ARGUMENTS, extract_table, get_engine


class Upserter:
    @classmethod
    async def create(
        cls,
        sa_obj: Union[sa.Table, DeclarativeMeta],
        remove_duplicates: bool = True,
        batch_by_columns_present: bool = True,
        on_dupe_update: Union[bool, List[str]] = True,
        column_name_filter_options: Optional[Dict[str, Any]] = None,
        column_value_filter_options: Optional[Dict[str, Any]] = None,
        engine: Optional[AsyncEngine] = None,
        logger: Logger = logger,
    ):
        """Upsert data rows to a postgresql database.

        Args:
            sa_obj (Union[sa.Table, DeclarativeMeta]): The SQLAlchemy table or entity corresponding to the database table that rows will be loaded to.
            remove_duplicates (bool, optional): Remove duplicates from upsert batches. Defaults to True.
            batch_by_columns_present (bool, optional): Group rows by columns present and execute upsert statement for each group. Defaults to True.
            on_dupe_update (Union[bool, List[str]]): List of columns that should be updated when primary key exists, or True for all columns, False for no columns. Defaults to True.
            column_name_filter_options (Optional[Dict[str, Any]], optional): Keyword arguments used to construct a `ColumnNameFilter` that will be applied to all rows before inserting. Defaults to None.
            column_value_filter_options (Optional[Dict[str, Any]], optional): Keyword arguments used to construct a `ColumnValueFilter` that will be applied to all rows before inserting. Defaults to None.
            engine (Optional[AsyncEngine], optional): The engine to use to execute sql statements. Defaults to None.
        """
        self = cls()
        self.logger = logger
        self.sa_obj = sa_obj
        self._table = extract_table(sa_obj)
        self.remove_duplicates = remove_duplicates
        self.batch_by_columns_present = batch_by_columns_present
        self.column_name_filter = (
            ColumnNameFilter(
                sa_obj=column_name_filter_options.pop("sa_obj", self.sa_obj),
                **column_name_filter_options,
            )
            if column_name_filter_options
            else None
        )
        self.column_value_filter = (
            ColumnValueFilter(
                sa_obj=column_value_filter_options.pop("sa_obj", self.sa_obj),
                **column_value_filter_options,
            )
            if column_value_filter_options
            else None
        )
        self.on_dupe_update = on_dupe_update
        if self.on_dupe_update == True:
            # update all columns that aren't primary key.
            self.on_dupe_update = [
                col_name
                for col_name, col in self._table.columns.items()
                if not col.primary_key
            ]
        elif self.on_dupe_update in (None, False):
            self.on_dupe_update = []
        self._primary_key_column_names = {
            c.name for c in self._table.primary_key.columns
        }
        self._n_primary_key_columns = len(self._primary_key_column_names)
        self.engine = engine or get_engine()
        async with self.engine.begin() as conn:
            # create table if it doesn't already exist.
            await create_table(conn, self._table)
        return self

    async def upsert(self, rows: List[Dict[str, Any]]) -> List[List[Dict[str, Any]]]:
        """Upsert rows to the database.

        Args:
            rows (List[Dict[str, Any]]): The rows to upsert.

        Returns:
            List[List[Dict[str, Any]]]: The row batches that were upserted.
        """
        if isinstance(rows, dict):
            rows = [rows]
        if self.column_name_filter:
            rows = [self.column_name_filter.filter_row(r) for r in rows]
        if self.remove_duplicates:
            rows = self._remove_duplicates(rows)
        if self.column_value_filter:
            rows = [self.column_value_filter.filter_row(r) for r in rows]
        row_groups = (
            self._groupby_columns(rows) if self.batch_by_columns_present else [rows]
        )
        async with self.engine.begin() as conn:
            for rows in row_groups:
                # check column of first row (all rows should have same columns)
                rows_batch_size = MAX_QUERY_ARGUMENTS // len(rows[0])
                for i in range(0, len(rows), rows_batch_size):
                    batch = rows[i : i + rows_batch_size]
                    statement = self._build_statement(batch)
                    self.logger.info(f"Loading {len(batch)} rows {self._table}")
                    await conn.execute(statement)
        return row_groups

    def _remove_duplicates(self, rows: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Remove rows that repeat a primary key.
        Multiple row in the same upsert statement can not have the same primary key.

        Args:
            rows (List[Dict[str, Any]]): Data row, possibly containing duplicates.

        Returns:
            List[Dict[str, Any]]: Duplicate-free data rows.
        """
        _keys = set()
        unique_key_rows = []
        for row in rows:
            key = tuple([row[col] for col in self._primary_key_column_names])
            if len(key) < self._n_primary_key_columns:
                logger.error(f"Row is missing primary key: {row}")
                continue
            if key not in _keys:
                unique_key_rows.append(row)
                _keys.add(key)
        if (unique_count := len(unique_key_rows)) < (row_count := len(rows)):
            logger.warning(
                f"{row_count - unique_count}/{row_count} rows had a duplicate primary key and will not be loaded."
            )
        return unique_key_rows

    def _groupby_columns(
        self, rows: List[Dict[str, Any]]
    ) -> List[List[Dict[str, Any]]]:
        """Group rows by column names present.
        We can not have rows with different columns in the same statement.

        Args:
            rows (List[Dict[str,Any]]): The rows to group.

        Returns:
            List[List[Dict[str, Any]]]: The grouped rows.
        """
        row_groups = defaultdict(list)
        for row in rows:
            row_groups[tuple(row.keys())].append(row)
        return list(row_groups.values())

    def _build_statement(self, rows: List[Dict[str, Any]]) -> Insert:
        """Construct an upsert statement to load `rows`.

        Args:
            rows (List[Dict[str,Any]]): The rows that will be upserted.

        Returns:
            Insert: An upsert statement.
        """
        statement = postgresql.insert(self._table).values(rows)
        if self.on_dupe_update and len(
            # check column of first row (all rows should have same columns)
            on_dupe_update := [c for c in self.on_dupe_update if c in rows[0]]
        ):
            statement = statement.on_conflict_do_update(
                index_elements=self._primary_key_column_names,
                set_={k: statement.excluded[k] for k in on_dupe_update},
            )
        else:
            statement = statement.on_conflict_do_nothing(
                index_elements=self._primary_key_column_names
            )
        return statement
