from datetime import datetime
from hashlib import md5
from logging import Logger
from operator import itemgetter
from typing import Any, Dict, List, Optional, Union
from uuid import uuid4

import sqlalchemy as sa
from logger_cfg import logger
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.dml import Insert
from sqlalchemy.orm.decl_api import DeclarativeMeta

from .types import DataIDIgnore
from .upsert import Upserter
from .utils import MAX_QUERY_ARGUMENTS, extract_table, get_typed_columns


class SourcesUpserter:
    @classmethod
    async def create(
        cls,
        source: str,
        upserter: Optional[Upserter] = None,
        logger: Logger = logger,
        **kwargs,
    ):
        """Upsert rows to a table and also upsert source metadata to a corresponding sources table.

        Args:
            source (str): The source that data rows are from.
            upserter (Upserter): The upserter that will be used to load rows.
            kwargs: If `upserter` argument is not provided, upserter kwargs may be provided and a new `Upserter` will be initialized.

        Raises:
            RuntimeError: If the table `upserter` is for does not contain the column `data_id` that will be used link tables.
        """
        self = cls()
        self.source = source
        self.upserter = upserter or await Upserter.create(logger=logger, **kwargs)
        self.logger = logger
        if "data_id" not in self.upserter._table.columns:
            raise RuntimeError(
                f"Table {self.upserter._table.name} must contain column `data_id` to be linked to a sources table."
            )
        self._sources_table = self.sources_table(self.upserter)
        async with self.upserter.engine.begin() as conn:
            await conn.run_sync(self._sources_table.create, checkfirst=True)

        # Names of all columns that should be used to create a data ID for a row.
        non_id_columns = {
            *get_typed_columns(DataIDIgnore, self.upserter.sa_obj),
            "data_id",
        }
        self._data_id_columns = [
            c.name for c in self.upserter._table.columns if c.name not in non_id_columns
        ]
        if not len(self._data_id_columns):
            raise ValueError(f"{self.upserter.sa_obj} does not have any ID columns.")

        return self

    @staticmethod
    def sources_table(sa_obj: Union[Upserter, sa.Table, DeclarativeMeta]) -> sa.Table:
        """Create a table to contain all sources of unique rows in a table.

        Args:
            sa_obj (Union[Upserter, sa.Table, DeclarativeMeta]): The entity to make the sources table for.

        Returns:
            sa.Table: The sources table.
        """
        table = sa_obj._table if isinstance(sa_obj, Upserter) else extract_table(sa_obj)
        _sources_table = sa.Table(
            f"{table.name}_sources",
            sa.MetaData(schema=table.schema),
            sa.Column(
                "data_id",
                sa.VARCHAR,
                sa.ForeignKey(table.c.data_id, ondelete="CASCADE"),
                primary_key=True,
            ),
            sa.Column("source", sa.VARCHAR, primary_key=True),
            sa.Column(
                "created",
                sa.DateTime(timezone=False),
                default=datetime.utcnow,
                nullable=False,
            ),
            sa.Column(
                "last_seen",
                sa.DateTime(timezone=False),
                default=datetime.utcnow,
                nullable=False,
            ),
            sa.Column("times_seen", sa.Integer, default=1, nullable=False),
        )
        return _sources_table

    async def upsert(self, rows: List[Dict[str, Any]]) -> None:
        """Upsert rows to the database and corresponding sources metadata to the sources table.

        Args:
            rows (List[Dict[str, Any]]): Data rows that should be loaded to the table.
        """
        if isinstance(rows, dict):
            rows = [rows]
        for row in rows:
            row["data_id"] = self._create_row_data_id(row)
        # load data rows to the database.
        row_groups = await self.upserter.upsert(rows)
        # load sources metadata to sources table.
        async with self.upserter.engine.begin() as conn:
            for rows in row_groups:
                rows_batch_size = MAX_QUERY_ARGUMENTS // len(rows[0])
                for i in range(0, len(rows), rows_batch_size):
                    batch = rows[i : i + rows_batch_size]
                    statement = self._build_statement(batch)
                    self.logger.info(
                        f"Loading {len(batch)} rows to sources table for source {self.source}."
                    )
                    await conn.execute(statement)

    def _create_row_data_id(self, row: Dict[str, Any]) -> str:
        """Generate an ID based on a hash of `row`'s data.

        Args:
            row (Dict[str, Any]): row with data that we should generate an ID for.

        Returns:
            str: The row's ID.
        """
        data_id_members = {k: v for k, v in row.items() if k in self._data_id_columns}
        if not len(data_id_members):
            _id = str(uuid4())
            self.logger.warning(
                "Row does not contain any fields that can be used to generate a data ID. Random ID '%s' will be used. (Data ID columns: %s. Row: %s)",
                _id,
                self._data_id_columns,
                row,
            )
            return _id
        # sort values by key so key ordering will not effect generated hash.
        data_id_members = [
            str(v).strip()
            for _, v in sorted(data_id_members.items(), key=itemgetter(0))
        ]
        return md5("".join(data_id_members).encode()).hexdigest()

    def _build_statement(self, rows: List[Dict[str, Any]]) -> Insert:
        """Construct an upsert statement to load source metadata for each row.

        Args:
            rows (List[Dict[str,Any]]): The rows to load source metadata for.

        Returns:
            Insert: An upsert statement.
        """
        statement = (
            postgresql.insert(self._sources_table)
            .values(
                [{"data_id": row["data_id"], "source": self.source} for row in rows]
            )
            .on_conflict_do_update(
                index_elements=["data_id", "source"],
                set_={
                    "last_seen": datetime.utcnow(),
                    "times_seen": sa.text(
                        f'{self._sources_table.schema}."{self._sources_table.name}".times_seen + 1'
                    ),
                },
            )
        )
        return statement
