import os
from functools import cache
from typing import Any, Set, Union, get_type_hints

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.orm.decl_api import DeclarativeMeta

# Number of query arguments. Useful for preventing batch inserts that are too large.
MAX_QUERY_ARGUMENTS = 17_000


@cache
def get_engine() -> AsyncEngine:
    """Create an SQLAlchemy engine and chache it so we only create it once.

    Returns:
        AsyncEngine: The sqlalchemy engine.
    """
    return create_async_engine(os.environ["DB_URL"])


def extract_table(sa_obj: Union[sa.Table, DeclarativeMeta]) -> sa.Table:
    """Extract the SQLAlchemy table from an entity, or return the passed argument if argument is already a table.

    Args:
        sa_obj (Union[sa.Table, DeclarativeMeta]): An entity or table object.

    Raises:
        ValueError: If argument is not an entity or table object.

    Returns:
        sa.Table: The table corresponding to the passed argument.
    """
    if isinstance(sa_obj, sa.Table):
        return sa_obj
    elif hasattr(sa_obj, "__table__"):
        return sa_obj.__table__
    raise ValueError(
        f"Object {sa_obj} is not an entity or table! Can not extract table."
    )


def get_typed_columns(
    of_type: Any, sa_obj: Union[sa.Table, DeclarativeMeta]
) -> Set[str]:
    """Find columns that are tagged as being of type `of_type` via querying type annotations and comments.

    Args:
        of_type (Any): The type who's type annotation should be searched for.
        sa_obj (Union[sa.Table, DeclarativeMeta]): The table or entity to query.

    Returns:
        Set[str]: Names of columns that are of type `of_type`.
    """
    table = extract_table(sa_obj)

    # extract type labels from column comments.
    type_columns = {
        col_name
        for col_name, col in table.columns.items()
        if col.comment == of_type.__name__
    }
    if not isinstance(sa_obj, sa.Table):
        # extract type labels from column type annotations.
        type_columns.update(
            {
                col_name
                for col_name, hint in get_type_hints(sa_obj).items()
                if hint is of_type
            }
        )
    return type_columns
